import logging
import os
from threading import Thread

import celery
from firexapp.broker_manager.broker_factory import RedisManager

from firex_flame.controller import FlameAppController
from firex_flame.event_file_processor import process_recording_file
from firex_flame.event_broker_processor import BrokerEventConsumerThread
from firex_flame.event_aggregator import FlameEventAggregator
from firex_flame.flame_helper import wait_until_path_exist

logger = logging.getLogger(__name__)


def create_broker_consumer_thread(broker_consumer_config, controller, event_aggregator,
                                  recording_file, shutdown_handler, logs_dir):
    try:
        broker_url = RedisManager.get_broker_url_from_logs_dir(logs_dir)
    except:
        logger.error(f"Failed to load broker URL from logs dir: {logs_dir}")
        raise
    else:
        celery_app = celery.Celery(broker=broker_url)
        return BrokerEventConsumerThread(celery_app,
                                         controller,
                                         event_aggregator,
                                         broker_consumer_config,
                                         recording_file,
                                         shutdown_handler)


def start_flame(webapp_port, broker_consumer_config, run_metadata, recording_file, shutdown_handler,
                extra_task_dump_paths, serve_logs_dir):
    event_aggregator = FlameEventAggregator()
    controller = FlameAppController(run_metadata, extra_task_dump_paths)
    if not recording_file or not os.path.isfile(recording_file):
        event_recv_thread = create_broker_consumer_thread(broker_consumer_config, controller, event_aggregator,
                                                          recording_file, shutdown_handler, run_metadata['logs_dir'])
        celery_app = event_recv_thread.celery_app
    else:
        event_recv_thread = Thread(target=process_recording_file, args=(event_aggregator, recording_file, run_metadata))
        celery_app = None
    event_recv_thread.start()

    if broker_consumer_config.receiver_ready_file:
        wait_until_path_exist(broker_consumer_config.receiver_ready_file, sleep_for=0.1)

    # Delaying of importing of all web dependencies is a deliberate startup performance optimization.
    # The broker should be listening for events as quickly as possible.
    from firex_flame.web_app import start_web_server
    return start_web_server(webapp_port, event_aggregator, run_metadata, controller, celery_app, serve_logs_dir)



"""
Process events from Celery in to flame data model.
"""

import logging
from pathlib import Path
import threading
import traceback
import json
import time

from celery.events import EventReceiver
from gevent import spawn, sleep
from gevent.queue import JoinableQueue

from firex_flame.controller import FlameAppController
from firex_flame.flame_helper import BrokerConsumerConfig

logger = logging.getLogger(__name__)


class RunningModelDumper:
    """
    Maintains a queue and a greenlet for async writing the flame data model. This prevents writing the model to disk
    from blocking the main celery event receiver.
    """

    SLIM_DUMP_TYPE = 'SLIM'
    TASK_DUMP_TYPE = 'TASK'
    STOP_DUMP_TYPE = 'STOP'
    ALL_WITHOUT_COMPLETED_DUMP_TYPE = 'ALL_WITHOUT_COMPLETED'
    ALL_DUMP_TYPES = {SLIM_DUMP_TYPE, TASK_DUMP_TYPE, STOP_DUMP_TYPE, ALL_WITHOUT_COMPLETED_DUMP_TYPE}

    def __init__(self,  flame_controller: FlameAppController, all_tasks_by_uuid):
        self.flame_controller = flame_controller

        # This is a JoinableQueue just to make testing easier. Clients will wait on the greenlet that processes
        # queue items, not the queue itself.
        self._queue = JoinableQueue()
        self.all_tasks_by_uuid = all_tasks_by_uuid

        # This class uses the 'task-completed' event type to loosely infer task completeness, since there is no
        # stronger indicator (e.g. task-failed does not mean completed, due to retries). This field stores task UUIDs
        # of tasks that have at one moment in time seen a task-completed event, which is distinct from tasks whose
        # latest event type is 'task-completed', since 'task-succeeded' can occur after 'task-completed'. Every event
        # after 'task-completed' will cause an additional write, since we don't know what can come after. This
        # is less than ideal.
        self.seen_task_completed_uuids = set()
        self._greenlet = spawn(self._consume_from_queue)

    def _dump_full_task(self, uuid, task):
        try:
            self.flame_controller.dump_full_task(uuid, task)
        except OSError as e:
            logger.warning(f"Failed to write {uuid} full task JSON: {e}")

    def _maybe_dump_task(self, uuid, event_type):
        if uuid in self.all_tasks_by_uuid:
            task = self.all_tasks_by_uuid[uuid]
            if (event_type in ['task-started-info', 'task-completed']
                    or uuid in self.seen_task_completed_uuids):
                self._dump_full_task(uuid, task)
                if event_type == 'task-completed':
                    self.seen_task_completed_uuids.add(uuid)
        else:
            logger.warning(f"Failed to write non-existant task with uuid: {uuid}")

    def _deduplicate_and_maybe_write_full_tasks(self, task_dump_work_items):
        deduplicated_task_uuid_to_event_type = {}
        for task_work_item in task_dump_work_items:
            uuid = task_work_item[1]
            event_type = task_work_item[2]
            assert uuid is not None, "Must have task UUID for TASK_DUMP_TYPE"
            assert event_type is not None, "Must have event type for TASK_DUMP_TYPE"

            # Never change away from task-completed, since we need to track if we've ever seen this type,
            # per task, to reduce total full task writes.
            if deduplicated_task_uuid_to_event_type.get(uuid) != 'task-completed':
                deduplicated_task_uuid_to_event_type[uuid] = event_type

        for uuid, event_type in deduplicated_task_uuid_to_event_type.items():
            self._maybe_dump_task(uuid, event_type)

    def _consume_from_queue(self):
        while True:
            triplet = self._queue.get()

            # drain queue and process all work items at once, de-duplicating work.
            work_items = [triplet] + [self._queue.get() for _ in range(len(self._queue))]
            work_item_types = {t[0] for t in work_items}

            try:
                if self.SLIM_DUMP_TYPE in work_item_types:
                    self.flame_controller.dump_slim_tasks(self.all_tasks_by_uuid)

                if self.TASK_DUMP_TYPE in work_item_types:
                    self._deduplicate_and_maybe_write_full_tasks([t for t in work_items if t[0] == self.TASK_DUMP_TYPE])

                if self.ALL_WITHOUT_COMPLETED_DUMP_TYPE in work_item_types:
                    all_uuids = set(self.all_tasks_by_uuid.keys())
                    uuids_without_completed = all_uuids.difference(self.seen_task_completed_uuids)
                    for uuid in uuids_without_completed:
                        self._dump_full_task(uuid, self.all_tasks_by_uuid[uuid])

                invalid_dump_types = work_item_types.difference(self.ALL_DUMP_TYPES)
                for invalid_dump_type in invalid_dump_types:
                    logger.warning(f"Unknown in-progress dumper work queue entry with type {invalid_dump_type}")

            except Exception as e:
                # TODO: narrow exception handling so that an error in handling of one dump_type doesn't fail others.
                logger.error("Failure while processing task-dumping work queue entry.")
                logger.exception(e)
            finally:
                for _ in range(len(work_items)):
                    self._queue.task_done()

                # Must be last.
                if self.STOP_DUMP_TYPE in work_item_types:
                    logger.debug(f"Stopping in progress model dumper.")
                    break

                # Let other greenlets run, possibly let work accumulate in the queue to allow work de-duplication
                sleep(0.2)

    def queue_write_slim(self):
        self._queue.put((self.SLIM_DUMP_TYPE, None, None))

    def queue_maybe_write_tasks(self, task_uuids_to_event_types):
        for task_uuid, event_type in task_uuids_to_event_types.items():
            self._queue.put((self.TASK_DUMP_TYPE, task_uuid, event_type))

    def write_remaining_and_wait_stop(self):
        self.queue_write_slim()
        self._queue.put((self.ALL_WITHOUT_COMPLETED_DUMP_TYPE, None, None))
        self._queue.put((self.STOP_DUMP_TYPE, None, None))
        self._greenlet.join() # Wait for queue to drain.

class BrokerEventConsumerThread(threading.Thread):
    """Events threading class
    """
    def __init__(self, celery_app, flame_controller: FlameAppController, event_aggregator, config: BrokerConsumerConfig,
                 recording_file: str, shutdown_handler):
        threading.Thread.__init__(self, daemon=True)
        self.celery_app = celery_app
        self.recording_file = recording_file
        self.flame_controller = flame_controller
        self.event_aggregator = event_aggregator
        self.max_try_interval = 2**config.max_retry_attempts if config.max_retry_attempts is not None else 32
        self.terminate_on_complete = config.terminate_on_complete
        self.stopped_externally = False
        self.shutdown_handler = shutdown_handler

        if config.receiver_ready_file:
            self.receiver_ready_file = Path(config.receiver_ready_file)
            assert not self.receiver_ready_file.exists(), \
                "Receiver ready file must not already exist: %s." % self.receiver_ready_file
        else:
            self.receiver_ready_file = None
        self.is_first_receive = True
        self.running_dumper_queue = RunningModelDumper(self.flame_controller, self.event_aggregator.tasks_by_uuid)

    def _create_receiver_ready_file(self):
        if self.is_first_receive:
            if self.receiver_ready_file:
                self.receiver_ready_file.touch()
            self.is_first_receive = False

    def _cleanup_tasks(self):
        # Create new events that change the run state of incomplete events.
        incomplete_task_events = self.event_aggregator.generate_incomplete_events()
        if incomplete_task_events:
            logger.warning("Forcing runstates of %d incomplete tasks to be terminal." % len(incomplete_task_events))
            self._aggregate_and_send(incomplete_task_events)
        else:
            logger.debug("All tasks already terminal following terminal root.")

    def run(self):
        self._run_from_broker()

    def _cleanup(self):
        try:
            self._cleanup_tasks()
            self.running_dumper_queue.write_remaining_and_wait_stop()
            self.flame_controller.dump_complete_data_model(self.event_aggregator)
        except Exception as e:
            logger.error("Failed to cleanup during receiver completion.")
            logger.exception(e)
        finally:
            logger.info("Completed receiver cleanup.")
            if self.terminate_on_complete and not self.stopped_externally:
                self.shutdown_handler.shutdown("Terminating on completion, as requested by input args.")

    def _run_from_broker(self):
        """Load the events from celery"""
        try:
            self._capture_events()
        finally:
            self._cleanup()

    def _capture_events(self):
        try_interval = 1
        while not self.event_aggregator.is_root_complete():
            try:
                try_interval *= 2
                with self.celery_app.connection() as conn:
                    conn.ensure_connection(max_retries=1, interval_start=0)
                    recv = EventReceiver(conn,
                                         handlers={"*": self._on_celery_event},
                                         app=self.celery_app)
                    try_interval = 1
                    self._create_receiver_ready_file()
                    recv.capture(limit=None, timeout=None, wakeup=True)
            except (KeyboardInterrupt, SystemExit) as e:
                self.stopped_externally = True
                self.shutdown_handler.shutdown(str(e))
            # pylint: disable=C0321
            except Exception:
                if self.event_aggregator.is_root_complete():
                    logger.info("Root task complete; stopping broker receiver (not entire server).")
                    return
                if self.shutdown_handler.shutdown_received:
                    self.stopped_externally = True
                    logger.info("Shutdown handler received shutdown request; stopping broker receiver.")
                    return
                logger.error(traceback.format_exc())
                if try_interval > self.max_try_interval:
                    logger.warning("Maximum broker retry attempts exceeded, stopping receiver (not entire server)."
                                   " Will no longer retry despite incomplete root task.")
                    return
                logger.debug("Try interval %d secs, still worth retrying." % try_interval)
                time.sleep(try_interval)

    def _on_celery_event(self, event):
        """Callback function for when an event is received

        Arguments:
            event(dict): The event to aggregate and send downstream.
        """
        # Append the event to the recording file if it is specified
        if self.recording_file:
            with open(self.recording_file, "a") as rec:
                event_line = json.dumps(event)
                rec.write(event_line + "\n")
        self._aggregate_and_send([event])

    def _aggregate_and_send(self, events):
        new_data_by_task_uuid = self.event_aggregator.aggregate_events(events)
        slim_update_data_by_uuid = self.flame_controller.send_sio_event(new_data_by_task_uuid,
                                                                        self.event_aggregator.tasks_by_uuid)
        if slim_update_data_by_uuid:
            self.running_dumper_queue.queue_write_slim()
        self.running_dumper_queue.queue_maybe_write_tasks({u: self.event_aggregator.tasks_by_uuid.get(u, {}).get('type')
                                                           for u in new_data_by_task_uuid.keys()})



import argparse
import os
import gzip
import json

from firex_flame.model_dumper import FlameModelDumper, find_flame_model_dir, get_model_full_tasks_by_names, \
    index_tasks_by_names, get_full_tasks_by_slim_pred
from firex_flame.event_aggregator import FlameEventAggregator, TASK_ARGS
from firex_flame.flame_helper import find_rec_file, find


def process_recording_file(event_aggregator: FlameEventAggregator, recording_file: str, run_metadata: dict):
    assert os.path.isfile(recording_file), "Recording file doesn't exist: %s" % recording_file

    real_rec = os.path.realpath(recording_file)
    if real_rec.endswith('.gz'):
        with gzip.open(real_rec, 'rt', encoding='utf-8') as rec:
            event_lines = rec.readlines()
    else:
        with open(recording_file) as rec:
            event_lines = rec.readlines()

    for event_line in event_lines:
        if not event_line:
            continue
        event = json.loads(event_line)
        event_aggregator.aggregate_events([event])

    if event_aggregator.is_root_complete():
        # Kludge incomplete runstates that will never become terminal.
        event_aggregator.aggregate_events(event_aggregator.generate_incomplete_events())

    if run_metadata.get('uid', None) is None and event_aggregator.root_uuid is not None:
        root_task = event_aggregator.tasks_by_uuid[event_aggregator.root_uuid]
        run_metadata['uid'] = find([TASK_ARGS, 'chain_args', 'uid'], root_task)
        run_metadata['chain'] = find([TASK_ARGS, 'chain'], root_task)


def dumper_main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--rec', help='Recording file to construct model from.')
    parser.add_argument('--dest_dir', help='Directory to which model should be dumped.')

    args = parser.parse_args()

    aggregator = FlameEventAggregator()
    process_recording_file(aggregator, args.rec, {})
    FlameModelDumper(root_model_dir=args.dest_dir).dump_aggregator_complete_data_model(aggregator)


def get_tasks_from_rec_file(log_dir=None, rec_filepath=None, mark_incomplete=False):
    assert bool(log_dir) ^ bool(rec_filepath), "Need exclusively either log directory of rec_file path."
    if not rec_filepath:
        rec_file = find_rec_file(log_dir)
    else:
        rec_file = rec_filepath
    assert os.path.exists(rec_file), "Recording file not found: %s" % rec_file
    aggregator = FlameEventAggregator()
    process_recording_file(aggregator, rec_file, {})

    if mark_incomplete:
        aggregator.aggregate_events(aggregator.generate_incomplete_events())

    return aggregator.tasks_by_uuid, aggregator.root_uuid


def get_model_or_rec_full_tasks_by_names(logs_dir, task_names):
    if os.path.isdir(find_flame_model_dir(logs_dir)):
        return get_model_full_tasks_by_names(logs_dir, task_names)

    rec_file = find_rec_file(logs_dir)
    if os.path.isfile(rec_file):
        tasks_by_uuid, _ = get_tasks_from_rec_file(rec_filepath=rec_file)
        return index_tasks_by_names(tasks_by_uuid.values(), task_names)

    raise Exception("Found neither model directory or rec_file, no source of task data in: %s" % logs_dir)


def get_model_or_rec_full_tasks_by_uuids(logs_dir, uuids):
    if os.path.isdir(find_flame_model_dir(logs_dir)):
        return get_full_tasks_by_slim_pred(logs_dir, lambda st: st['uuid'] in uuids)

    rec_file = find_rec_file(logs_dir)
    if os.path.isfile(rec_file):
        tasks_by_uuid, _ = get_tasks_from_rec_file(rec_filepath=rec_file)
        return {u: t for u, t in tasks_by_uuid.items() if u in uuids}

    raise Exception("Found neither model directory or rec_file, no source of task data in: %s" % logs_dir)


def get_model_or_rec_full_task(logs_dir, uuid):
    task_by_uuid = get_model_or_rec_full_tasks_by_uuids(logs_dir, [uuid])
    return task_by_uuid[uuid]



# Need fastentrypoints to monkey patch setuptools for faster console_scripts
# noinspection PyUnresolvedReferences
import fastentrypoints
from setuptools import setup, find_packages
import versioneer

setup(name='firex_flame',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='FireX event processor and web server.',
      url='https://github.com/FireXStuff/firex-flame',
      author='Core FireX Team',
      author_email='firex-dev@gmail.com',
      license='BSD-3-Clause',
      packages=find_packages(),
      zip_safe=True,
      install_requires=[
            "Flask==1.0.2",
            "Flask-AutoIndex==0.6.4",
            "Werkzeug==0.16.1",
            "python-socketio==4.1.0",
            "python-engineio==3.10.0", # 3.11.0 causes launch crash when tornado dependency isn't present.
            'firexapp',
            # Middle UI version number indicates expected server API version.
            # e.g 0.20.x means this server adheres to contract version 0.20 between UI and server.
            "firex_flame_ui<0.27",
            "requests",
            "beautifulsoup4",
            "paramiko",
            "gevent-websocket",
            "gevent",
            # Without pinning, importlib 3.0.0 causes:
            #     IsADirectoryError: [Errno 21] Is a directory: '/<path>/site-packages/firex_flame_ui'
            "importlib-metadata==1.6.0",
            "importlib-resources==1.0.2",
      ],
      package_data={
        'firex_flame': ['templates/*.html'],
      },
      entry_points={
          'console_scripts': ['firex_flame = firex_flame.__main__:main',
                              'flame_dump = firex_flame.event_file_processor:dumper_main'],
          'firex_tracking_service': ['flame_launcher = firex_flame.launcher:FlameLauncher', ],
      },)



import logging

from firex_flame.event_aggregator import slim_tasks_by_uuid
from firex_flame.model_dumper import FlameModelDumper
from firex_flame.flame_helper import query_partial_tasks, get_dict_json_md5

logger = logging.getLogger(__name__)


def _send_listening_query_sio_event(sio_server, new_data_by_task_uuid, tasks_by_uuid,
                                    listening_query_config_hashes_to_sids):
    for query_hash, query_and_clients in listening_query_config_hashes_to_sids.items():
        # Only bother filtering/emitting the event if there are clients for this query.
        if query_and_clients['client_sids']:
            # FIXME: if the new data contains additional children,
            queried_task_data = query_partial_tasks(new_data_by_task_uuid.keys(),
                                                    query_and_clients['query_config'],
                                                    tasks_by_uuid)
            if queried_task_data:
                sio_server.emit('tasks-query-update', data=queried_task_data, room=query_hash)


class FlameAppController:

    def __init__(self, run_metadata, extra_task_representations):
        self.run_metadata = run_metadata
        self.model_dumper = FlameModelDumper(firex_logs_dir=self.run_metadata['logs_dir'])
        self.extra_task_representations = extra_task_representations
        self.listening_query_config_hashes_to_sids = {}

        self.sio_server = None  # Set after creation.

    def send_sio_event(self, new_data_by_task_uuid, tasks_by_uuid):
        slim_update_data_by_uuid = {uuid: task
                                    for uuid, task in slim_tasks_by_uuid(new_data_by_task_uuid).items()
                                    if task}
        # sio_server can be lazy initialized. Since the event receiving process starts before the
        # web modules are loaded, extremely early events can't be delivered.
        if self.sio_server:
            # Avoid sending events if there aren't fields the downstream cares about.
            if slim_update_data_by_uuid:
                self.sio_server.emit('tasks-update', slim_update_data_by_uuid)
            _send_listening_query_sio_event(self.sio_server, new_data_by_task_uuid, tasks_by_uuid,
                                            self.listening_query_config_hashes_to_sids)
        return slim_update_data_by_uuid

    def dump_initial_metadata(self):
        self.model_dumper.dump_metadata(self.run_metadata, root_complete=False, flame_complete=False)

    def dump_complete_data_model(self, event_aggregator):
        # Even broker's running_dumper_queue will dump task JSONs, so no need to dump again.
        dump_task_jsons = False
        self.model_dumper.dump_aggregator_complete_data_model(event_aggregator, self.run_metadata,
                                                              self.extra_task_representations, dump_task_jsons)

    def dump_full_task(self, uuid, task):
        self.model_dumper.dump_full_task(uuid, task)

    def dump_slim_tasks(self, all_tasks_by_uuid):
        self.model_dumper.dump_slim_tasks(all_tasks_by_uuid)

    def add_client_task_query_config(self, sid, query_config):
        config_md5 = get_dict_json_md5(query_config)
        if config_md5 not in self.listening_query_config_hashes_to_sids:
            self.listening_query_config_hashes_to_sids[config_md5] = {
                'query_config': query_config,
                'client_sids': set(),
            }
        self.listening_query_config_hashes_to_sids[config_md5]['client_sids'].add(sid)
        self.sio_server.enter_room(sid, room=config_md5)

    def remove_client_task_query(self, sid):
        for config_md5, config in self.listening_query_config_hashes_to_sids.items():
            if sid in config['client_sids']:
                config['client_sids'].remove(sid)
                self.sio_server.leave_room(sid, room=config_md5)
